/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#include "createwordlist.hpp"

CreateWordlist::CreateWordlist(Settings &se)
    : settings(se)
    , configDir(settings.getString("options/configDir", ""))
    , infile(settings.getString("wordlist/create/filename", "../cwtl").c_str())
    , tagfile((configDir + settings.getString("wordlist/tagfile", "tagfile")).c_str(), std::ios::out|std::ios::trunc) {
    if (!infile) {
        std::cerr << "Error: Can't open input file. Aborting..." << std::endl;
        std::exit(1);
    }
    outfile = gzopen((configDir + settings.getString("wordlist/wordfile", "outfile.gz")).c_str(), "wb9");
    if (!outfile) {
        std::cerr << "Error: Can't create output file. Aborting..." << std::endl;
        std::exit(1);
    }
    if (!tagfile) {
        std::cerr << "Error: Can't create output file. Aborting..." << std::endl;
        std::exit(1);
    }

    std::cerr << "Reading input data";
    readFromFile();
    std::cerr << ".done\n";

    std::cerr << "Creating data structure";
    transformData();
    std::cerr << ".done\n";

    std::cerr << "Writing to files";
    writeToFile();
    std::cerr << ".done\n";
}

void CreateWordlist::readFromFile() {
    boost::match_results<std::string::const_iterator> result;
    boost::regex exp;
    exp.set_expression(settings.getString("wordlist/create/expression", ".+\\t(.+)\\t(.+)\\t(.+)"), boost::regbase::normal);
    std::string line;
    int tagid = 0;
    int counter = 0;
    try {
        while (std::getline(infile, line)) {

            if (!regex_match(line, result, exp)) {
                throw InputParseException("Error while reading wordlist from file.");
            }
            if (result.size() != 4) {
                throw InputParseException("Wrong number of tokens read.");
            }
            std::string word = result[1];
            std::string tag = result[2];
            std::string lemma = result[3];

            Utils::lowercase(word);
            Utils::lowercase(tag);
            Utils::lowercase(lemma);
	       
            //transform(word.begin(), word.end(), word.begin(), tolower);
            //transform(tag.begin(), tag.end(), tag.begin(), tolower);
            //transform(lemma.begin(), lemma.end(), lemma.begin(), tolower);
            tagid = tagmap.insert(tag);
            wlt.insert(std::make_pair(word, std::make_pair(lemma, tagid)));
            lwt.insert(std::make_pair(lemma, std::make_pair(word, tagid)));
            if (++counter % 3500 == 0) {
                std::cerr << ".";
            }
        }
    }
    catch (const InputParseException &e) {
        std::cerr << e.what() << std::endl;
        std::exit(1);
    }
}

void CreateWordlist::transformData() {
    int counter = 0;
    for (WLT::const_iterator iter = wlt.begin(); iter != wlt.end(); ++iter) {
        Corpus::WordToken wt[100];
        int wtcount = 0;
	  
        // Tag, Lemma, Word
        wt[wtcount].set(iter->second.second, iter->second.first, iter->first);
        wtcount += 1 + wt[wtcount].size();
        for (int i = 0; i < wtcount; i++) {
            tokenvector.push_back(wt[i]);
        }
        if (++counter % 7000 == 0) {
            std::cerr << ".";
        }
    }
    for (WLT::const_iterator iter = lwt.begin(); iter != lwt.end(); ++iter) {
        Corpus::WordToken wt[100];
        int wtcount = 0;
	  
        // Tag, Word, Lemma
        wt[wtcount].set(iter->second.second, iter->second.first, iter->first);
        wtcount += 1 + wt[wtcount].size();
        for (int i = 0; i < wtcount; i++) {
            tokenvector.push_back(wt[i]);
        }
        if (++counter % 7000 == 0) {
            std::cerr << ".";
        }
    }
}

void CreateWordlist::writeToFile() {
    int counter = 0;
    tagfile << tagmap;
    int length = tokenvector.size() / 2;
    gzwrite(outfile, &length, sizeof(length));
    for (int i = 0; i < length * 2; i++) {
        gzwrite(outfile, &tokenvector[i], sizeof(Corpus::WordToken));
        if (++counter % 10000 == 0) {
            std::cerr << ".";
        }
    }
    gzclose(outfile);
}

void printUsage(const char *name) {
    std::cerr << "Usage:  " << name << " [ options... ]\n"
              << "Options:\n"
              << "   -f<filename>\tSpecifies which settings file to use\n"
              << "   -c<dir>\tSpecifies the config directory\n"
              << "   -w<filename>\tSpecifies the input filename\n"
              << "   -h\t\tThis help text\n\n"
              << "The input and output filenames are specified in the\n"
              << "'wordlist' section of the settings file.\n\n";
    std::exit(0);
}

int main(int argc, char* argv[]) {
    try {
        XMLPlatformUtils::Initialize();
    }
    catch (const XMLException& toCatch) {
        char* message = XMLString::transcode(toCatch.getMessage());
        std::cout << "Error during initialization! :\n"
                  << message << "\n";
        delete [] message;
        std::exit(1);
    }
    std::string settingsfile = "../conf/settings.xml";

    for (int i = 1; argv[i]; ++i) {
        char *arg = argv[i];
        if (*arg == '-') {
            switch (*++arg) {
            case 'f': 
                if (*++arg == '\0') printUsage(argv[0]); 
                settingsfile = arg; break;
            case 'c':
            case 'w': break;
            case 'h':
            default: printUsage(argv[0]); break;
            }
        }
    }
    
    // This needs to be in a new scope, because the destructor of the BufferedSettings must run before XMLPlatformUtils::Terminate();
    {
      BufferedSettings se(settingsfile);
      if (!se.insertCommandLineParameters(argc, argv)) {
	printUsage(argv[0]);
      }
      CreateWordlist cwl(se);
    }

    XMLPlatformUtils::Terminate();
    return 0;
}
