/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#include "damerau.hpp"

Damerau::Damerau(Wordlist &wl, Settings &se, ErrorWriter &ew)
     : wordlist(wl)
     , settings(se)
     , errorwriter(ew)
     , letters("abcdefghijklmnopqrstuvwxyz")
     , confusionMatrix(letters.length(), letters.length())
     , confusionSum(letters.length())    
     , generator(settings.getInt("options/seed", 123456789))
    /*, dst(0, std::numeric_limits<int>::max())*/
     , random_number(generator, distribution_type(0,std::numeric_limits<int>::max())) 
     , allowExistingWords(settings.getBool("damerau/allowExistingWords", true))
     , forceAllowWords(settings.getBool("damerau/forceAllowWords", false))
     , allowTagChange(settings.getBool("damerau/allowTagChange", true))
     , forceAllowTag(settings.getBool("damerau/forceAllowTag", false)) 
     , reportDamerau(settings.getString("damerau/reportName", "damerau")) {
     numLetters = letters.length();
     rnd = NULL;
     int subst = settings.getInt("damerau/subst", 1);
     int ins = settings.getInt("damerau/ins", 1);
     int del = settings.getInt("damerau/del", 1);
     int transp = settings.getInt("damerau/transp", 1);
     setProbabilities(subst, ins, del, transp);
     std::string configDir = settings.getString("options/configDir", "");
     readConfusionMatrix(configDir + settings.getString("damerau/confusionMatrix", "confusion.any"));
     //std::cerr << "alphabet is '" << letters << "'\n";
}

Damerau::~Damerau() {
    delete rnd;
}

void Damerau::setProbabilities(const char s, const char i, const char d, const char t) {
     int sum;
     subst = s;
     ins = i + subst;
     del = d + ins;
     transp = t + del;
     sum = transp;
     delete rnd;     
     /*dst = distribution_type(0,sum-1);*/
     rnd = new generator_type(generator, distribution_type(0,sum-1));
}

bool Damerau::apply(std::list<WordTag>& listwt) const {
     // Skip "blank" lines
     if (listwt.front().blank()) {
	  return false;
     }
     try {
	  WtStrInt ret = apply(listwt.front());
	  WordTag newword = ret.get<0>();
	  std::string reason = ret.get<1>();
	  int idx = ret.get<2>();
	  /* Check whether a change was done */
	  if (newword.word() != listwt.front().word()) {
	       /* Allow only if word exists and existing words are allowed,
		  or if word doesn't exist and existing words are not obligatory. */
	       if (wordlist.lookup(newword.word()) && allowExistingWords ||
		   !wordlist.lookup(newword.word()) && !(allowExistingWords && forceAllowWords)) {
		    /* Allow only if tag has changed and tag change is allowed,
		       or if tag didn't change and tag change is not obligatory*/
		    if (newword.tag() != listwt.front().tag() && allowTagChange ||
			newword.tag() == listwt.front().tag() && !(allowTagChange && forceAllowTag)) {
			 newword.setWord(wordlist.preserveCase(listwt.front().word(), newword.word()));
			 newword.addDesc(listwt.front().desc());
			 newword.addDesc(reportDamerau);
			 errorwriter.insertDamerau(reason, idx, listwt.front(), newword);
			 listwt.pop_front();
			 listwt.push_front(newword);
			 return true;
		    }
	       }
	  }
     }
     catch (const StrangeLetterException &) {
	  //std::cerr << "Strange letter...\n";
     }
     //std::cerr << "\n";
     return false;
}

Damerau::WtStrInt Damerau::apply(const WordTag &wt) const {
     std::string word = wt.word();
     std::string tag = wordlist.tag(wt.tag());
     std::string reason = "unknown";
     int idx;
     if (wt.word().length() <= 0) {
	  std::cerr << "Note: Empty word.\n";
	  return boost::make_tuple(WordTag("", wordlist.untag(wordlist.unknownTag()), wt.row(), wt.origWord(), wt.origTag(), wt.orig()), reason, idx);
     }
     int num = (*rnd)();
     if (num < subst) {     // substitution
	  idx = random_number() % word.length();
	  if (idx == 0) {
	       idx = random_number() % word.length();
	  }
	  word[idx] = confuseLetter(word[idx]);
	  reason = "subst";
     }
     else if (num < ins) {  // insertion
	  idx = random_number() % (word.length() + 1);
	  if (idx == 0) {
	       idx = random_number() % (word.length() + 1);
	  }
	  char mem;
	  if (idx == 0) {
	       mem = confuseLetter(word[0]);
	  }
	  else {
	       mem = confuseLetter(word[idx - 1]);
	  }
	  word.insert(idx, 1, mem);
	  reason = "ins";
     }
     else if (num < del) {  // deletion
	  if (word.length() <= 1) {
	       return boost::make_tuple(wt, reason, idx);
	  }
	  idx = random_number() % word.length();
	  if (idx == 0) {
	       idx = random_number() % word.length();
	  }
	  word.erase(idx, 1);
	  reason = "del";
     }
     else {                 // transposition
	  if (word.length() <= 1) {
	       return boost::make_tuple(wt, reason, idx);
	  }
	  idx = random_number() % (word.length() - 1);
	  if (idx == 0) {
	       idx = random_number() % (word.length() - 1);
	  }
	  char swap = word[idx];
	  word[idx] = word[idx + 1];
	  word[idx + 1] = swap;
	  reason = "transp";
     }
     WordTag newwordtag(word, wordlist.getTag(word, wt.tag()), wt.row(), wt.origWord(), wt.origTag(), wt.orig());
     return boost::make_tuple(newwordtag, reason, idx);
}

char Damerau::confuseLetter(const char in) const {
     int index;
     for (index = 0; index < numLetters; index++) {
	  if (letters[index] == in) {
	       break;
	  }
     }
     if (index == numLetters) {
	  //std::cerr << "Warning: Strange letter found: '" << in << "'\n";
	  throw StrangeLetterException("Strange lettter found.");
     }
     int value = random_number() % (confusionSum[index] - 1) + 1;
     for (int j = 0; j < numLetters; j++) {
	  if (confusionMatrix[index][j] >= value) {
	       //std::cout << "(confusion: '" << in << "' -> '" << letters[j] << "') ";
	       return letters[j];
	  }
     }
     return '#';
}

void Damerau::readConfusionMatrix(const std::string& filename) {
     std::ifstream file(filename.c_str());
     if (!file) {
	  std::cerr << "Warning: Can't open confusion matrix  file '" 
		    << filename << "'! Using default...\n";
     }
     else {
	  std::string str;
	  int row = 0;
	  try {
	       // Read the alphabet on the first line of the file
	       letters = "";
	       numLetters = 0;
	       std::getline(file, str);
	       boost::tokenizer<> firstline(str);
	       boost::tokenizer<>::iterator firstiter = firstline.begin();
	       for (/*empty*/; firstiter != firstline.end(); ++firstiter) {
		    letters += *firstiter;
	       }
	       numLetters = letters.length();

	       // Read the matrix from file
	       while (std::getline(file, str)) {
		    int col = 0;
		    boost::tokenizer<> tok(str);
		    boost::tokenizer<>::iterator beg = tok.begin();
		    confusionMatrix[row][col] = boost::lexical_cast<int>(*beg);
		    confusionSum[row] += confusionMatrix[row][col];
		    ++beg;
		    col++;
		    for(/*empty*/; beg != tok.end(); ++beg){
			 confusionMatrix[row][col] = confusionMatrix[row][col - 1] + boost::lexical_cast<int>(*beg);
			 col++;
		    }
		    if (col != numLetters) {
			 throw ConfusionMatrixException("Wrong number of columns in " + filename);
		    }
		    confusionSum[row] = confusionMatrix[row][col - 1];
		    row++;
	       }
	       if (row != numLetters) {
		    throw ConfusionMatrixException("Wrong number of lines in" + filename);
	       }
	       return;
	  }
	  catch (boost::bad_lexical_cast &) {
	       std::cerr << "Warning: Not a numerical value (0-255) in '" << filename 
			 << "'. Using default confusion matrix..\n";
	  }
	  catch (ConfusionMatrixException &e) {
	       std::cerr << "Warning: " << e.what()
			 << ". Using default confusion matrix..\n";
	  }
	  catch (...) {
	       std::cerr << "Error: Unexpected exception. da1\n";
	  }
     }

     // Default values for confusion matrix
     letters = "abcdefghijklmnopqrstuvwxyz";
     numLetters = letters.length();
     const int default_value = 1;
     for (int i = 0; i < numLetters; i++) {
	  confusionMatrix[i][0] = (i == 0) ? 0 : default_value;
	  for (int j = 1; j < i; j++) {
	       confusionMatrix[i][j] = confusionMatrix[i][j - 1] + default_value;
	  }
	  confusionMatrix[i][i] = confusionMatrix[i][i - 1];
	  for (int j = i + 1; j < numLetters; j++) {
	       confusionMatrix[i][j] = confusionMatrix[i][j - 1] + default_value;
	  }
	  confusionSum[i] = (numLetters - 1) * default_value;
     }
}

