/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#ifndef DYNAMIC_H
#define DYNAMIC_H

namespace Corpus
{
     template<class T>
     T max(T t1, T t2) { return t1 < t2 ? t2 : t1; }

     template<class T> class Integer;
     template<int SIZE> class Str;
    
     typedef Integer<signed char> Char;
     typedef Integer<short> Short;
     typedef Integer<int> Int;
     typedef Str<8> String;

     template<class T1, class T2> class Dynamic2;
     template<class T1, class T2, class T3> class Dynamic3;

     class WordToken;
     class TagName;

     template<class T>
     T *next(int size, const void *p) {
	  return (T *)((char *)p + size);
     }
}

template<class T>
class Corpus::Integer
{ 
public:
     typedef T type;

     Integer() : data(0) {}

     T get(int len, int size) const {
	  if(len == 0)
	       return data;
	  else
	       return next<Integer<T> >(size, this)->data;
     }
     int set(T t, int size, int len = 0) {
	  if(len == 0)
	       data = t;
	  else {
	       data = -len;
	       next<Integer<T> >(size, this)->set(t, size, 0);
	  }
	  return 0;
     }
     int size() const { return data >= 0 ? 0 : -data; };

private:
public:
     T data;
};

template<int SIZE>
class Corpus::Str   
{ 
public:
     char data[SIZE]; 

public:
     Str() {
	  for(int i = 0; i < SIZE; i++)
	       data[i] = '\0';
     }

     typedef std::string type; 
     type get(int len, int size) const {
	  return copy() + (len == 0 ? "" : get_next(len, size));
     }
     int set(type t, int size, int len = 0) {
	  if(copy(t, len))
	       return next<Str>(size, this)->set(t, size, len + 1);
	  else
	       return len;
     }
     int size() const { return 1; };

private:
     type copy() const {
	  type s;
	  for(int i = 0; i < SIZE && data[i]; i++)
	       s.append(1, data[i]);
	  return s;
     }
     bool copy(type s, int len) {
	  int i;
	  int j = 0;
	  for(i = len * SIZE; j < SIZE && s[i]; i++, j++)
	       data[j] = s[i];
	  return s[i] != '\0';
     }
     type get_next(int len, int size) const {
	  return next<Str>(size, this)->get(len - 1, size);
     }
};

template<class T1, class T2>
class Corpus::Dynamic2
{
public:
     int size() const { return t1.size(); }
     typename T1::type get_t1() const { return t1.get(size(), sizeof *this); }
     typename T2::type get_t2() const { return t2.get(size(), sizeof *this); }

     void set(typename T1::type d1, typename T2::type d2) {
	  int i2 = t2.set(d2, sizeof *this);
	  int i1 = i2;
	  t1.set(d1, sizeof *this, i1);
     }

     int parse(const std::string &s) {
	  std::istringstream istr(s);
	  
	  typename T1::type t1;
	  typename T2::type t2;
	  istr >> t1 >> t2;
	  
	  set(t1, t2);

	  return size() + 1;
     }

private:
public:
     T1 t1;
     T2 t2;
};

template<class T1, class T2, class T3>
class Corpus::Dynamic3
{
public:
     int      size() const   { return t1.size(); }
     typename T1::type get_t1() const { return t1.get(size(), sizeof *this); }
     typename T2::type get_t2() const { return t2.get(size(), sizeof *this); }
     typename T3::type get_t3() const { return t3.get(size(), sizeof *this); }

     void set(typename T1::type d1, typename T2::type d2, typename T3::type d3) {
	  int i2 = t2.set(d2, sizeof *this);
	  int i3 = t3.set(d3, sizeof *this);
	  int i1 = max(i2, i3);
	  t1.set(d1, sizeof *this, i1);
     }

     int parse(const std::string &s) {
	  std::istringstream istr(s);
	  
	  typename T1::type t1;
	  typename T2::type t2;
	  typename T3::type t3;
	  istr >> t1 >> t2 >> t3;
	  
	  set(t1, t2, t3);

	  return size() + 1;
     }
     
private:
public:
     T1 t1;
     T2 t2;
     T3 t3;
};

class Corpus::WordToken :
     public Corpus::Dynamic3<Corpus::Int, Corpus::String, Corpus::String> {
public:
     WordToken() {}
     WordToken(int i, std::string l, std::string w) {
	  set(i, l, w);
     }

     int         tag()   const  { return get_t1(); }
     std::string lemma() const  { return get_t2(); }
     std::string word()  const  { return get_t3(); }
     int         data()  const  { return t1.data; }
};

class Corpus::TagName :
     public Corpus::Dynamic2<Corpus::Int, Corpus::String> {
public:
     TagName() {}
     TagName(int i, std::string l) {
	  set(i, l);
     }

     int         tag()  const { return get_t1(); }
     std::string name() const { return get_t2(); }
};

#endif
