/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#include <iostream>
#include <cstdlib>

#include <xercesc/util/PlatformUtils.hpp>

#include "wordlist.hpp"
#include "bufferedsettings.hpp"
#include "wordtag.hpp"

#include "sounderror.hpp"
#include "inputreader.hpp"
#include "errorintroducer.hpp"
#include "splitcompound.hpp"
#include "damerau.hpp"
#include "driver.hpp"

XERCES_CPP_NAMESPACE_USE


#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <strings.h>
#include <string.h>

// Assuming the text is Swedish text, this should return 1 if the
// encoding is UTF-8, and it should return 0 if we find something that
// looks like latin-1
int looksLikeUTF(const char *s) {
  if(s == NULL) {
    return 0;
  }

  int last = 1;
  for(int i = 0; s[i] != '\0'; i++) {
    if(s[i] < 0 && last > 0) {
      if(s[i] == -27
	 || s[i] == -28
	 || s[i] == -10
	 || s[i] == -59
	 || s[i] == -60
	 || s[i] == -42) {
	return 0;
      }
    }
    if(s[i] < 0 && last == -61) {
      return 1;
    }
    last = s[i];
  }
  
  return 0;
}

/* Converts UTF-8 to Latin-1 correctly if we assume the string
   contents are Swedish text. Based on code example from Mark Ransom
   
   returns new size of buffer */
int utf2latin1(char *s, int len) {
  
  int n = len;
  if(n <= 0) {
    return 0;
  }

  char *tempBuf = new char[n+1];

  unsigned int codepoint = 0;

  int j = 0;
  for (int i = 0; s[i] != 0; ) {
    unsigned char c = (unsigned char) (s[i]);
    if (c <= 0x7f)
      codepoint = c;
    else if (c <= 0xbf)
      codepoint = (codepoint << 6) | (c & 0x3f);
    else if (c <= 0xdf)
      codepoint = c & 0x1f;
    else if (c <= 0xef)
      codepoint = c & 0x0f;
    else
      codepoint = c & 0x07;

    i++;
    
    if (((s[i] & 0xc0) != 0x80) && (codepoint <= 0x10ffff)) {
      if (codepoint <= 255) {
	tempBuf[j++] = (char)(codepoint);
      } else {
	// do whatever you want for out-of-bounds characters
	tempBuf[j++] = 'X'; // Store something away so we can convert back later? TODO: Jonas
      }
    }
  }

  for(int i = 0; i < j; i++) {
    s[i] = tempBuf[i];
  }
  s[j] = 0;

  delete [] tempBuf;

  return j;
}

const int SOCKETBUFSIZE = 1000000;
int read_data(int socket, unsigned char *buf) {
  /* char buf[SOCKETBUFSIZE+1]; */
  int resultCount = 0;
  int done = 0;
  int len = 0, offs = 0;
  unsigned char *end = 0;
  
  buf[0] = 0;
  buf[SOCKETBUFSIZE] = 0;

  while(offs < SOCKETBUFSIZE && !done) {
    len = read(socket, buf+offs, SOCKETBUFSIZE - offs);

    if(len > 0) {
      buf[SOCKETBUFSIZE] = 0;

      end = (unsigned char *)strstr((char *)buf, "ENDQ");
      if(end != NULL) {
	*end = 0;
	done = 1;
      } else {
	offs += len;
      }
      
    } else {
      buf[offs] = 0;
      end = buf + offs;
      done = 1;
    }

    if(offs >= SOCKETBUFSIZE) {
      buf[SOCKETBUFSIZE-1] = 0;
      end = &(buf[SOCKETBUFSIZE-1]);
    }
  }

  if(looksLikeUTF((char *)buf)) {
    int newSize = utf2latin1((char *)buf, end - buf);
    end = buf + newSize;
  }
  
  return resultCount;
}

// taken from a socket primer
int establish(int portnum)
{ 
  const int MAXHOSTNAME = 1000;
  char myname[MAXHOSTNAME+1];
  int s;
  struct sockaddr_in sa;
  struct hostent *hp;
  
  bzero(&sa,sizeof(struct sockaddr_in));
  gethostname(myname,MAXHOSTNAME);
  hp= gethostbyname(myname);
  if (hp == NULL)
    return -1;
  sa.sin_family= hp->h_addrtype; /* this is our host address */
  sa.sin_port= htons(portnum); /* this is our port number */
  if ((s = socket(AF_INET,SOCK_STREAM,0)) < 0)
    return -1;
  if (bind(s,(struct sockaddr *)&sa,sizeof(sa)) < 0) {
    close(s);
    return -1;
  }
  listen(s, 5); /* max # of queued connects */
  return s;
}

// taken from a socket primer
int get_connection(int s // socket
		   )
{ 
  struct sockaddr_in isa;
  socklen_t i;
  int t;
  
  i = sizeof(isa);
  getsockname(s,(struct sockaddr *)&isa,&i);
  
  if ((t = accept(s,(struct sockaddr *)&isa,&i)) < 0)   /* accept connection if there is one */
    return -1;
  return t;
}


void startupXML() {
  try {        
    XMLPlatformUtils::Initialize();
  }
  catch (const XMLException& toCatch) {
    char* message = XMLString::transcode(toCatch.getMessage());
    std::cout << "Error during initialization! :\n"
	      << message << "\n";
    delete [] message;
    std::exit(1);
  }
}

void shutdownXML() {
  XMLPlatformUtils::Terminate();
}

void printUsage(const char *name) {
  std::cerr << "Usage:  " << name << " [ options... ]\n"
	    << "Options:\n"
	    << "   -p<port number>\tSpecifies which port to listen to\n"
	    << "   -t<temp folder>\tSpecifies where to put temp files\n"
	    << "   -f<filename>\tSpecifies which settings file to use\n"
    //	    << "   -i<filename>\tSpecifies the input filename\n"
    //	    << "   -o<filename>\tSpecifies the output filename\n"
    //	    << "   -x<filename>\tSpecifies the name of the error XML description\n"
	    << "   -n<seed>\tSpecifies the seed for the random number generator\n"
	    << "   -d<rate>\tEnable the Damerau module\n\t\tSpecifying <rate> overrides default error rate\n"
	    << "   -s<rate>\tEnable the SplitCompound module\n\t\tSpecifying <rate> overrides default error rate\n"
	    << "   -e<rate>\tEnable the Sound Error module\n\t\tSpecifying <rate> overrides default error rate\n"
	    << "   +d\t\tDisable the Damerau module\n"
	    << "   +s\t\tDisable the SplitCompound module\n"
	    << "   +e\t\tDisable the Sound Error module\n"
	    << "   -h\t\tThis help text\n";
  std::exit(0);
}

int main(int argc, char* argv[]) {
  std::string settingsfile = "settings.xml";

  std::string tempFolder = "/var/tmp";
  int portNo = -1;

  int server_socket = 0;
  int client_socket = 0;
  unsigned char socketBuf[SOCKETBUFSIZE+1];

  int pid = -1;
  std::string tempIn;
  std::string tempOut;
  std::string tempX;
  std::string tempFile = "";
  std::ofstream temp;
  std::ifstream res;
  std::ifstream resx;
  std::stringstream stream;
  std::string toSend;
  std::ostringstream os;
  
  /* handler for child processes */
  signal(SIGCHLD, SIG_IGN);
  
  for (int i = 1; argv[i]; ++i) {
    char *arg = argv[i];
    if (*arg == '-') {
      switch (*++arg) {
      case 'f': 
	if (*++arg == '\0')
	  printUsage(argv[0]); 
	settingsfile = arg;
	break;
      case 'i':
      case 'o':
      case 'x':
      case 'n':
      case 'd':
      case 's':
      case 'e': 
      case '-': break;
      case 'p': 
	if (*++arg == '\0')
	  printUsage(argv[0]);
	else 
	  portNo = atoi(arg);
	break;
      case 't': 
	if (*++arg == '\0')
	  printUsage(argv[0]);
	else 
	  tempFolder = arg;
	break;
      case 'h':
      default: printUsage(argv[0]); break;
      }
    }
  }

  if(portNo <= 0) {
    printUsage(argv[0]);
  } else {


    if ((server_socket = establish(portNo)) < 0) {
      fprintf(stderr, "%s: could not open socket for listening\n", argv[0]);
      exit(1);
    } 


    while(1) {
      fprintf(stderr, "%s: waiting for input on port %d\n", argv[0], portNo);
      client_socket = 0;
      if ((client_socket = get_connection(server_socket)) < 0) {
	if (errno == EINTR)
	  continue;
	fprintf(stderr, "%s: error on accept\n", argv[0]);
	exit(1);
      }
    
      // fork all client calls, that way, if they hang or crash,
      // the server keeps working
    
      switch(fork()) {
      case -1 : // error
	fprintf(stderr, "%s: error on fork()\n", argv[0]);
	exit(1);
      case 0 : // child
	close(server_socket);
	alarm(60*2); // if we still haven't finished in 2 minutes, die...

	// read data from socket
	read_data(client_socket, socketBuf);
	alarm(0); // remove alarm

	// get our process ID to create hopefully unique filenames for tempfiles
	pid = getpid();

	tempFile = tempFolder;
	if(tempFile[tempFile.size() - 1] != '/') {
	  tempFile += "/";
	}
	tempFile += "sokcetmissplel";
	os << pid;
	tempFile += os.str();

	tempIn = tempFile;
	tempIn += ".in";
	tempOut = tempFile;
	tempOut += ".out";
	tempX = tempFile;
	tempX += ".x";

	// make sure the temp folder exists
	system((std::string("mkdir -p ") + tempFolder).c_str());

	// dump all the data from the socket to a file
	temp.open(tempIn);
	temp << socketBuf;
	temp.close();

	// ------------------------------------------------------------------
	// Here is the original Missplel code 
	// ------------------------------------------------------------------
	startupXML();
    
	{
	  BufferedSettings settings(settingsfile);
	  if (!settings.insertCommandLineParameters(argc, argv)) {        
	    printUsage(argv[0]);
	  }    
	  if (settings.getString("options/configDir", "") != "") {
	    settings.addParameter("options/configDir", settings.getString("options/configDir", "../conf") + "/");
	  }

	  // -----------------------------------------------
	  // replace the filenames with our temp files
	  // -----------------------------------------------
	  settings.addParameter("errorWriter/filename", tempX);
	  settings.addParameter("output/filename", tempOut);
	  settings.addParameter("input/filename", tempIn);
	  
	  Wordlist wordlist(settings);
	  ErrorWriter errorwriter(wordlist, settings);
	  errorwriter.insertInfo(argv, settings.getString("options/seed", "123456789"), settingsfile);

	  InputReader inputreader(wordlist, settings);

	  IntroducerList introducerlist(wordlist, settings, errorwriter, inputreader);
	  SoundError sounderror(wordlist, settings, errorwriter);
	  SplitCompound splitCompound(wordlist, settings, errorwriter, inputreader);
	  Damerau damerau(wordlist, settings, errorwriter);

	  Driver driver(wordlist, settings, inputreader, damerau, splitCompound, introducerlist, sounderror);
	  driver.createErrors();

	  errorwriter.write();
	}
	shutdownXML();

	// ------------------------------------------------------------------
	// End of the original Missplel code 
	// ------------------------------------------------------------------

	// open the two output files from Missplel and write the contents to the socket
	res.open(tempOut);
	resx.open(tempX);
	stream << resx.rdbuf() << "\n<output>\n" << res.rdbuf() << "\n</output>\n";
	toSend = stream.str();

	write(client_socket, toSend.c_str(), toSend.size() * sizeof(char));

	res.close();
	resx.close();

	// remove the temp files
	std::remove(tempOut.c_str());
	std::remove(tempIn.c_str());
	std::remove(tempX.c_str());
	
	close(client_socket);
	exit(0); // child is finished
      default : // parent
	close(client_socket);
	wait(NULL); // To avoid spamming lots of processes, wait for child to finish before accepting new connections.
      }
    } // while(1)
    
  }
    
  std::cout << "\nGoodbye.\n";
  return 0;
}
