/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#ifndef WORDLIST_H
#define WORDLIST_H

#include <string>
#include <map>
#include <fstream>
#include <sstream>
#include <iostream>
#include <algorithm>
#include <cctype>

#include "settings.hpp"
#include "dynamic.hpp"
#include "exceptions.hpp"
#include "wordlist_iterator.hpp"

typedef std::map<int, std::string> MapIntStr;
typedef std::map<std::string, int> MapStrInt;

/**
 * The class Wordlist represents a wordlist (duh!)
 * The stored information contains: word, lemma and tag
 */
class Wordlist {
public:
     Wordlist(Settings&);
     ~Wordlist();

     typedef WordlistIterator const_iterator;

     /* Iterators to the "word part" of the wordlist */
     const_iterator begin();
     const_iterator end();

     /* Iterators to the "lemma part" of the wordlist */
     const_iterator lemma_begin();
     const_iterator lemma_end();

     const_iterator findFirstWord(const std::string&);
     const_iterator findFirstLemma(const std::string&);
     
     const_iterator findLastWord(const std::string&);
     const_iterator findLastLemma(const std::string&);

     const std::string findLemma(const std::string&, int);
     const std::string findWord(const std::string&, int);

     bool lookup(const std::string&);

     /* Convert tag number to tag string and vice versa */
     const std::string& tag(int) const;
     int untag(std::string);

     const std::string& unknownTag() const;
     const std::string& unknownLemma() const;

     /* Get a tag for a word, and choose the specified tag if possible */
     int getTag(const std::string&, const int);

     std::string preserveCase(const std::string&, const std::string&) const;
     bool allUpper(const std::string&) const;
     char toUpper(char) const;
     char toLower(char) const;
private:
     const Corpus::WordToken* nextToken(const Corpus::WordToken*) const;
     inline int nextToken(const int, const Corpus::WordToken*) const;

     const Corpus::WordToken* findFirst(const std::string&, const Corpus::WordToken*);
     const Corpus::WordToken* findLast(const std::string&, const Corpus::WordToken*);

     int startOfWord(int, const Corpus::WordToken*) const;
     int startOfNextWord(int, const Corpus::WordToken*) const;

     /* This is an ugly hack to prevent the search routine to get stuck in
      * an infinite loop. When searching through the "fast file", it is possible
      * and highly likely that the binary search will find a cell not containing
      * the beginning of a word. To find the beginning of the word the functions
      * startOfWord and startOfNextWord are used. But if only one of these are
      * used it is possible the binary search will loop infinitely.
      * Why? If the min limit is 1 and max limit is 3 the next position in the
      * search will be 2. But if this is in the middle of a word and always get
      * corrected to 3 (the beginning of the nex word), we will have an infinite
      * loop.
      * For that reason the to functions are called every other time.
      */
     int startOfSomeWord(int num, const Corpus::WordToken wlt[]) {
	  useNext = !useNext;
	  if (useNext) {
	       return startOfNextWord(num, wlt);
	  }
	  return startOfWord(num, wlt);
     }

     Settings &settings;
     Corpus::WordToken *wltlong;
     Corpus::WordToken *lwtlong;
     int wtcount;
     MapIntStr tags;
     MapStrInt untags;
     bool useNext;
     std::string unknowntag;
     std::string unknownlemma;
};

#endif
